/***********************************************************************
*
* asmpass2.c - Pass 2 for the TI 990 assembler.
*
* Changes:
*   05/21/03   DGP   Original.
*   06/15/03   DGP   Added registerarg function.
*   06/23/03   DGP   Added date/time and creator to EOF record.
*   07/01/03   DGP   Fixed TEXT/BYTE processing.
*                    Fixed negative REF/DEF equ'd values.
*   07/02/03   DGP   Change hanging byte methods.
*                    Fixed register expressions.
*   07/08/03   DGP   Moved get date/time to always get it.
*   07/10/03   DGP   Fixed EVEN to use absolute for new org.
*   07/17/03   DGP   Fixed BSS & BES hanging byte.
*	
***********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <time.h>

#include "asmdef.h"

extern int pc;
extern int rorgpc;
extern int symbolcount;
extern int errcount;
extern int errnum;
extern int absolute;
extern int indorg;
extern int p1errcnt;
extern int pgmlength;
extern char errline[10][120];
extern char inbuf[MAXLINE];
extern SymNode *symbols[MAXSYMBOLS];
extern ErrTable p1error[MAXERRORS];

char ttlbuf[TTLSIZE+2];
char idtbuf[IDTSIZE+2];

static int linecnt = MAXLINE;
static int curp1err = 0;
static int objrecnum = 0;
static int objcnt = 0;
static int linenum = 0;
static int pagenum = 0;
static int printed = FALSE;
static int listing = TRUE;
static int listmode;
static char datebuf[48];
static char pcbuf[6];
static char objbuf[16];
static char objrec[82];
static char opbuf[32];
static char cursym[32];
static char lstbuf[MAXLINE];
static struct tm *timeblk;

static uint16 byteacc;
static int bytecnt;

/***********************************************************************
* printheader - Print header on listing.
***********************************************************************/

static void
printheader (FILE *lstfd)
{
   if (linecnt > LINESPAGE)
   {
      linecnt = 0;
      if (pagenum) fputc ('\f', lstfd);
      fprintf (lstfd, H1FORMAT,
	       VERSION,
	       idtbuf,
	       datebuf,
	       ++pagenum);
      fprintf (lstfd, H2FORMAT, ttlbuf); 
   }
}

/***********************************************************************
* printline - Print line on listing.
*
*0        1         2         3         4
*1234567890123456789012345678901234567890
*
* oooo  FFFF  FFFF  FFFF  LLLL  TTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTT
* oooo  FFFF  FFFF        LLLL  TTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTT
* oooo  FFFF              LLLL  TTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTT
*
***********************************************************************/

static void
printline (FILE *lstfd)
{
   printheader (lstfd);
   fprintf (lstfd, L1FORMAT, pcbuf, opbuf,
	    linenum, lstbuf);
   linecnt++;
}

static void
printdata (FILE *lstfd, int opc)
{
   printheader (lstfd);
   sprintf (pcbuf, PCFORMAT, opc);
   fprintf (lstfd, L2FORMAT, pcbuf, opbuf);
   linecnt++;
}

/***********************************************************************
* printsymbols - Print the symbol table.
***********************************************************************/

static void
printsymbols (FILE *lstfd)
{
   int i, j;
   char type;

   linecnt = MAXLINE;
   j = 0;

   for (i = 0; i < symbolcount; i++)
   {
      printheader (lstfd);

      if (!(symbols[i]->xop))
      {
	 if (symbols[i]->global) type = 'G';
	 else if (symbols[i]->external) type = 'E';
	 else if (symbols[i]->relocatable) type = '\'';
	 else type = ' ';

	 fprintf (lstfd, SYMFORMAT,
		  symbols[i]->symbol,
		  symbols[i]->value & 0xFFFF,
		  type);
	 j++;
	 if (j >= 5)
	 {
	    fprintf (lstfd, "\n");
	    linecnt++;
	    j = 0;
	 }
      }
   }
   fprintf (lstfd, "\n");
}

/***********************************************************************
* punchfinish - Punch a record with sequence numbers.
***********************************************************************/

static void
punchfinish (FILE *outfd)
{
   if (objcnt)
   {
      strncpy (&objrec[objcnt], "80000F", 6);
      sprintf (&objrec[SEQUENCENUM], SEQFORMAT, ++objrecnum);
      fputs (objrec, outfd);
      memset (objrec, ' ', sizeof(objrec));
      objcnt = 0;
   }
}

/***********************************************************************
* punchrecord - Punch an object value into record.
***********************************************************************/

static void 
punchrecord (FILE *outfd)
{
   int len = strlen (objbuf);

   if (!indorg)
   {
      if (objcnt+len >= CHARSPERREC)
      {
	 punchfinish (outfd);
      }
      strncpy (&objrec[objcnt], objbuf, len);
      objcnt += len;
      objbuf[0] = '\0';
   }
}

/***********************************************************************
* punchsymbols - Punch REF and DEF symbols.
***********************************************************************/

static void
punchsymbols (FILE *outfd)
{
   int i;

   for (i = 0; i < symbolcount; i++)
   {
      if (symbols[i]->external)
      {
	 sprintf (objbuf, REFFORMAT,
	       symbols[i]->relocatable ? RELEXTRN_TAG : ABSEXTRN_TAG,
	       symbols[i]->value & 0xFFFF, symbols[i]->symbol);
         
         punchrecord (outfd);
      }
      else if (symbols[i]->global)
      {
	 sprintf (objbuf, DEFFORMAT,
	       symbols[i]->relocatable ? RELGLOBAL_TAG : ABSGLOBAL_TAG,
	       symbols[i]->value & 0xFFFF, symbols[i]->symbol);
         punchrecord (outfd);
      }
   }
}

/***********************************************************************
* puncheof - Punch EOF mark.
***********************************************************************/

static void
puncheof (FILE *outfd)
{
   char temp[80];

   punchfinish (outfd);
   objrec[0] = EOFSYM;
   sprintf (temp, "%-8.8s  %02d/%02d/%02d  %02d:%02d:%02d    ASM990 %s",
	    idtbuf,
	    timeblk->tm_mon, timeblk->tm_mday, timeblk->tm_year - 100,
	    timeblk->tm_hour, timeblk->tm_min, timeblk->tm_sec,
	    VERSION);
   strncpy (&objrec[7], temp, strlen(temp));
   sprintf (&objrec[SEQUENCENUM], SEQFORMAT, ++objrecnum);
   fputs (objrec, outfd);
}

/***********************************************************************
* generalarg - Process general operand arg.
***********************************************************************/

static char *
generalarg (char *bp, int *ts, int *tr, int *saddr, int *srelo, int *spc,
char *term, int pcinc)
{
   char *tok;
   int toktyp;
   int laddr = -1;
   int lr = 0;
   int lrelo = 0;
   int val;
   int lpc = 0;
   char lterm;

   lpc = *spc;
   *ts = 0;
   if (*bp == MEMSYM)
   {
      bp++;
      *ts = 0x2;
      lpc += 2;
      bp = exprscan (bp, &laddr, &lterm, &lrelo, 2, FALSE, pcinc);
      laddr = laddr & 0xFFFF;
      if (lterm == '(')
      {
         bp = exprscan (bp, &lr, &lterm, &val, 2, FALSE, 0);
	 if (lterm == ')') lterm = *bp++;
      }
   }
   else
   {
      if (*bp == '*')
      {
	 char *cp;

	 bp++;
	 *ts = 0x1;
	 cp = bp;
	 while (!isspace(*cp) && *cp != ',') cp++;
	 if (*(cp-1) == '+')
	 {
	    char tterm;
	    *ts = 0x3;
	    *(cp-1) = '\0';
	    bp = exprscan (bp, &lr, &lterm, &val, 2, FALSE, 0);
	    lterm = *cp++;
	    bp = cp;
	 }
	 else
	    bp = exprscan (bp, &lr, &lterm, &val, 2, FALSE, 0);
      }
      else
	 bp = exprscan (bp, &lr, &lterm, &val, 2, FALSE, 0);
   }

   if (lr < 0 || lr > 15)
   {
      errcount++;
      sprintf (errline[errnum++], "Invalid register value: %d", lr);
      lr = 0;
   }

   *saddr = laddr;
   *tr = lr;
   *srelo = lrelo;
   *spc = lpc;
   *term = lterm;

   return (bp);
}

/***********************************************************************
* registerarg - Process register operand arg.
***********************************************************************/

static char *
registerarg (char *bp, int *reg, char *term)
{
   int rd;
   int junk;
   char lt;

   if (*bp == MEMSYM)
   {
	 errcount++;
	 sprintf (errline[errnum++], "Not a register: %s", bp);
	 rd = 0;
   }
   else
   {

      bp = exprscan (bp, &rd, &lt, &junk, 2, FALSE, 0);
      if (rd < 0 || rd > 15)
      {
	 errcount++;
	 sprintf (errline[errnum++], "Invalid register value: %d", rd);
	 rd = 0;
      }
   }
   *reg = rd;
   *term = lt;
   return (bp);
}

/***********************************************************************
* p21op - Process type 1 operation code.
***********************************************************************/

static void
p21op (OpCode *op, int flag, char *bp, FILE *outfd)
{
   char *cp;
   char *tok;
   int td = 0;
   int ts = 0;
   int rd = 0;
   int rs = 0;
   int daddr = -1;
   int saddr = -1;
   int toktyp;
   int srelo;
   int drelo;
   int val;
   int spc;
   char term;

   sprintf (pcbuf, PCFORMAT, pc);
   spc = 0;
   while (isspace(*bp)) bp++;
   bp = generalarg (bp, &ts, &rs, &saddr, &srelo, &spc, &term, spc + 2);
   if (term == ',')
   {
      bp = generalarg (bp, &td, &rd, &daddr, &drelo, &spc, &term, spc + 2);
   }
   val = op->opvalue | (td << 10) | (rd << 6) | (ts << 4) | rs;
   sprintf (objbuf, OBJFORMAT, ABSDATA_TAG, val);
   punchrecord (outfd);

   if (saddr < 0 && daddr < 0)
   {
      sprintf (opbuf, OP1FORMAT, val);
   }
   else if (saddr >= 0 && daddr >= 0)
   {
      sprintf (opbuf, OP3FORMAT, val,
	       saddr, srelo ? '\'' : ' ',
	       daddr, drelo ? '\'' : ' ');
      sprintf (objbuf, OBJFORMAT, srelo ? RELDATA_TAG : ABSDATA_TAG, saddr);
      punchrecord (outfd);
      sprintf (objbuf, OBJFORMAT, drelo ? RELDATA_TAG : ABSDATA_TAG, daddr);
      punchrecord (outfd);
   }
   else if (saddr < 0)
   {
      sprintf (opbuf, OP2FORMAT, val, daddr, drelo ? '\'' : ' ');
      sprintf (objbuf, OBJFORMAT, drelo ? RELDATA_TAG : ABSDATA_TAG, daddr);
      punchrecord (outfd);
   }
   else 
   {
      sprintf (opbuf, OP2FORMAT, val, saddr, srelo ? '\'' : ' ');
      sprintf (objbuf, OBJFORMAT, srelo ? RELDATA_TAG : ABSDATA_TAG, saddr);
      punchrecord (outfd);
   }
   pc += 2 + spc;
}

/***********************************************************************
* p22op - Process type 2 operation code.
***********************************************************************/

static void
p22op (OpCode *op, int flag, char *bp, FILE *outfd)
{
   int addr = 0;
   int val;
   int relocatable;
   int disp;
   char term;

   sprintf (pcbuf, PCFORMAT, pc);

   bp = exprscan (bp, &addr, &term, &relocatable, 2, FALSE, 0);
   if (op->opmod == 0) /* A Jump instruction */
   {
      if (*bp == MEMSYM)
      {
	 errcount++;
	 sprintf (errline[errnum++], "Not a valid jump operand: %s", bp);
	 disp = 0;
      }
      else
      {
	 disp = addr/2 - (pc+2)/2;
	 if (disp < -128 || disp > 128)
	 {
	       errcount++;
	       sprintf (errline[errnum++], "Jump displacement too large: %d",
			disp);
	       disp = 0;
	 }
      }
   }
   else	/* A CRU instruction */
   {
      disp = addr;
   }
   val = op->opvalue | (disp & 0xFF);
   sprintf (opbuf, OP1FORMAT, val);
   sprintf (objbuf, OBJFORMAT, ABSDATA_TAG, val);
   pc += 2;
}

/***********************************************************************
* p23op - Process type 3 operation code.
***********************************************************************/

static void
p23op (OpCode *op, int flag, char *bp, FILE *outfd)
{
   char *cp;
   char *tok;
   int ts = 0;
   int rs = 0;
   int rd = 0;
   int saddr = -1;
   int toktyp;
   int srelo;
   int val;
   int spc;
   char term;

   sprintf (pcbuf, PCFORMAT, pc);
   spc = 0;
   while (isspace(*bp)) bp++;
   bp = generalarg (bp, &ts, &rs, &saddr, &srelo, &spc, &term, 2);
   if (term == ',')
   {
      bp = registerarg (bp, &rd, &term);
   }

   val = op->opvalue | (rd << 6) | (ts << 4) | rs;
   sprintf (objbuf, OBJFORMAT, ABSDATA_TAG, val);
   punchrecord (outfd);

   if (saddr < 0)
   {
      sprintf (opbuf, OP1FORMAT, val);
   }
   else 
   {
      sprintf (opbuf, OP2FORMAT, val, saddr, srelo ? '\'' : ' ');
      sprintf (objbuf, OBJFORMAT, srelo ? RELDATA_TAG : ABSDATA_TAG, saddr);
      punchrecord (outfd);
   }
   pc += 2 + spc;
}

/***********************************************************************
* p24op - Process type 4 operation code.
***********************************************************************/

static void
p24op (OpCode *op, int flag, char *bp, FILE *outfd)
{
   char *cp;
   char *tok;
   int ts = 0;
   int rs = 0;
   int count = 0;
   int saddr = -1;
   int toktyp;
   int srelo;
   int val;
   int spc;
   char term;

   sprintf (pcbuf, PCFORMAT, pc);
   spc = 0;
   while (isspace(*bp)) bp++;
   bp = generalarg (bp, &ts, &rs, &saddr, &srelo, &spc, &term, 2);
   if (term == ',')
   {
      if (*bp == MEMSYM)
      {
	 errcount++;
	 sprintf (errline[errnum++], "Not a count: %s", bp);
	 count = 0;
      }
      else
      {
	 bp = exprscan (bp, &count, &term, &val, 2, FALSE, 0);
	 if (count < 0 || count > 15)
	 {
	    errcount++;
	    sprintf (errline[errnum++], "Invalid count value: %d", count);
	    count = 0;
	 }
      }
   }
   val = op->opvalue | (count << 6) | (ts << 4) | rs;
   sprintf (objbuf, OBJFORMAT, ABSDATA_TAG, val);
   punchrecord (outfd);

   if (saddr < 0)
   {
      sprintf (opbuf, OP1FORMAT, val);
   }
   else 
   {
      sprintf (opbuf, OP2FORMAT, val, saddr, srelo ? '\'' : ' ');
      sprintf (objbuf, OBJFORMAT, srelo ? RELDATA_TAG : ABSDATA_TAG, saddr);
      punchrecord (outfd);
   }
   pc += 2 + spc;
}

/***********************************************************************
* p25op - Process type 5 operation code.
***********************************************************************/

static void
p25op (OpCode *op, int flag, char *bp, FILE *outfd)
{
   int r = 0;
   int c = 0;
   int val;
   char term;

   sprintf (pcbuf, PCFORMAT, pc);
   bp = registerarg (bp, &r, &term);
   if (term == ',')
   {
      if (*bp == MEMSYM)
      {
	 errcount++;
	 sprintf (errline[errnum++], "Not a shift count: %s", bp);
	 c = 0;
      }
      else
      {
	 bp = exprscan (bp, &c, &term, &val, 2, FALSE, 0);
	 if (c < 0 || c > 15)
	 {
	    errcount++;
	    sprintf (errline[errnum++], "Invalid shift count: %d", c);
	    c = 0;
	 }
      }
   }

   val = op->opvalue | (c << 4) | r;
   sprintf (opbuf, OP1FORMAT, val);
   sprintf (objbuf, OBJFORMAT, ABSDATA_TAG, val);
   pc += 2;
}

/***********************************************************************
* p26op - Process type 6 operation code. 
***********************************************************************/

static void
p26op (OpCode *op, int flag, char *bp, FILE *outfd)
{
   char *cp;
   char *tok;
   int ts = 0;
   int rs = 0;
   int saddr = -1;
   int toktyp;
   int srelo;
   int val;
   int spc;
   char term;

   sprintf (pcbuf, PCFORMAT, pc);
   spc = 0;
   while (isspace(*bp)) bp++;
   bp = generalarg (bp, &ts, &rs, &saddr, &srelo, &spc, &term, 2);

   val = op->opvalue | (ts << 4) | rs;
   sprintf (objbuf, OBJFORMAT, ABSDATA_TAG, val);
   punchrecord (outfd);

   if (saddr < 0)
   {
      sprintf (opbuf, OP1FORMAT, val);
   }
   else 
   {
      sprintf (opbuf, OP2FORMAT, val, saddr, srelo ? '\'' : ' ');
      sprintf (objbuf, OBJFORMAT, srelo ? RELDATA_TAG : ABSDATA_TAG, saddr);
      punchrecord (outfd);
   }
   pc += 2 + spc;
}

/***********************************************************************
* p27op - Process type 7 operation code. 
***********************************************************************/

static void
p27op (OpCode *op, int flag, char *bp, FILE *outfd)
{
   char *tok;
   int mod = 0;
   int toktyp;
   char term;
   int tag = 0;

   sprintf (pcbuf, PCFORMAT, pc);
   sprintf (opbuf, OP1FORMAT, op->opvalue);
   sprintf (objbuf, OBJFORMAT, ABSDATA_TAG, op->opvalue);
   pc += 2;
}

/***********************************************************************
* p28op - Process type 8 operation code. 
***********************************************************************/

static void
p28op (OpCode *op, int flag, char *bp, FILE *outfd)
{
   int r = 0;
   int i = 0;
   int relo;
   int val;
   char term;

   sprintf (pcbuf, PCFORMAT, pc);
   val = op->opvalue;

   if (op->opmod == 1) /* Instructions without register operand */
   {
      sprintf (objbuf, OBJFORMAT, ABSDATA_TAG, val);
      goto NOREG;
   }

   bp = registerarg (bp, &r, &term);
   val = op->opvalue | r;
   sprintf (objbuf, OBJFORMAT, ABSDATA_TAG, val);

   if (op->opmod == 2) /* Instructions with only register operand */
   {
      sprintf (opbuf, OP1FORMAT, val);
      pc += 2;
      return;
   }

   if (term == ',')
   {
NOREG:
      bp = exprscan (bp, &i, &term, &relo, 2, FALSE, 2);
      if (i < -65634 || i > 65535)
      {
	 errcount++;
	 sprintf (errline[errnum++], "Invalid immediate value: %d", i);
	 i = 0;
      }
      i &= 0xFFFF;
   }
   punchrecord (outfd);
   sprintf (opbuf, OP2FORMAT, val, i, relo ? '\'' : ' ');
   sprintf (objbuf, OBJFORMAT, relo ? RELDATA_TAG : ABSDATA_TAG, i);
   pc += 4;
}

/***********************************************************************
* p29op - Process type 9 operation code. 
***********************************************************************/

static void
p29op (OpCode *op, int flag, char *bp, FILE *outfd)
{
   char *cp;
   char *tok;
   int ts = 0;
   int rs = 0;
   int rd = 0;
   int saddr = -1;
   int toktyp;
   int srelo;
   int val;
   int spc;
   char term;

   sprintf (pcbuf, PCFORMAT, pc);
   spc = 0;
   while (isspace(*bp)) bp++;
   bp = generalarg (bp, &ts, &rs, &saddr, &srelo, &spc, &term, 2);
   if (term == ',')
   {
      bp = registerarg (bp, &rd, &term);
   }

   val = op->opvalue | (rd << 6) | (ts << 4) | rs;
   sprintf (objbuf, OBJFORMAT, ABSDATA_TAG, val);
   punchrecord (outfd);

   if (saddr < 0)
   {
      sprintf (opbuf, OP1FORMAT, val);
   }
   else 
   {
      sprintf (opbuf, OP2FORMAT, val, saddr, srelo ? '\'' : ' ');
      sprintf (objbuf, OBJFORMAT, srelo ? RELDATA_TAG : ABSDATA_TAG, saddr);
      punchrecord (outfd);
   }
   pc += 2 + spc;
}

/***********************************************************************
* p2pop - Process Pseudo operation code.
***********************************************************************/

static int
p2pop (OpCode *op, char *bp, FILE *lstfd, FILE *outfd)
{
   SymNode *s;
   char *cp;
   char *tok;
   int i, j, k, l;
   int opc;
   int val;
   int oval;
   int relocatable;
   int toktyp;
   char term;
   char temp[10];

   sprintf (pcbuf, PCFORMAT, pc);

   /*
   ** If word based op and "hanging" byte, punch it.
   */

   switch (op->opvalue)
   {
   case AORG_T:
   case BES_T:
   case BSS_T:
   case DATA_T:
   case DORG_T:
   case END_T:
   case EVEN_T:
   case NOP_T:
   case RT_T:
      if (bytecnt)
      {
	 byteacc = (byteacc << 8) | 0;
	 sprintf (objbuf, OBJFORMAT, ABSDATA_TAG, byteacc);
	 punchrecord (outfd);
	 bytecnt = 0;
      }
      break;
   default: ;
   }

   switch (op->opvalue)
   {

   case AORG_T:
      if (!absolute && !indorg) rorgpc = pc;
      cp = exprscan (bp, &val, &term, &relocatable, 2, FALSE, 0);
      if (val < 0 || val > 65535)
      {
	 errcount++;
	 sprintf (errline[errnum++], "Invalid AORG value: %d", val);
	 val = 0;
      }
      indorg = FALSE;
      absolute = TRUE;
      pc = val;
      sprintf (pcbuf, PCFORMAT, pc);
      sprintf (objbuf, OBJFORMAT, ABSORG_TAG, val);
      break;

   case BES_T:
   case BSS_T:
      bp = exprscan (bp, &val, &term, &relocatable, 1, FALSE, 0);
      if (val < 0 || val > 65535)
      {
	 errcount++;
	 sprintf (errline[errnum++], "Invalid storage value: %s", tok);
	 val = 0;
      }
      pc += val;
      sprintf (objbuf, OBJFORMAT, absolute ? ABSORG_TAG : RELORG_TAG, pc);
      break;

   case BYTE_T:
      i = j = k = l = 0;
      opc = pc;
      do {
	 bp = exprscan (bp, &val, &term, &relocatable, 1, FALSE, 0);
	 if (val < -256 || val > 255)
	 {
	    errcount++;
	    sprintf (errline[errnum++], "Invalid BYTE value: %d", val);
	 }
	 val &= 0xFF;

	 byteacc = byteacc << 8 | val;
	 bytecnt++;
	 sprintf (temp, CHAR1FORMAT, val);
	 strncpy (&opbuf[j], temp, 2);
	 j += 2;
	 pc++;
	 i++;
	 if (bytecnt == 2)
	 {
	    sprintf (objbuf, OBJFORMAT, ABSDATA_TAG, byteacc);
	    punchrecord (outfd);
	    bytecnt = 0;
	    byteacc = 0;
	 }
	 if (i == 2)
	 {
	    i = 0;
	    k++;
	    if (k == 3)
	    {
	       printed = TRUE;
	       if (l++ == 0)
	       {
		  if (listmode && listing)
		  {
		     printline (lstfd);
		  }
		  opc = pc;
	       }
	       else
	       {
		  if (listmode && listing)
		  {
		     printdata (lstfd, opc);
		  }
		  opc = pc;
	       }
	       strcpy (opbuf, OPBLANKS);
	       j = k = 0;
	    }
	    else
	    {
	       j += 2;
	    }
	 }
      
      } while (term == ',');

      if (k || i)
      {
	 if (printed)
	 {
	    printdata (lstfd, opc);
	 }
      }
      break;

   case DATA_T:
      opc = pc;
      if (pc & 0x0001) pc++;
      sprintf (pcbuf, PCFORMAT, pc);
      if (pc != opc)
      {
	 sprintf (objbuf, OBJFORMAT, absolute ? ABSORG_TAG : RELORG_TAG, pc);
	 punchrecord (outfd);
      }
      opc = pc;
      i = j = 0;
      do {
	 bp = exprscan (bp, &val, &term, &relocatable, 2, FALSE, 0);
	 if (val < -65534 || val > 65535)
	 {
	    errcount++;
	    sprintf (errline[errnum++], "Invalid DATA value: %d", val);
	 }
	 val &= 0xFFFF;
	 sprintf (temp, DATA1FORMAT, val, relocatable ? '\'' : ' ');
	 strncpy (&opbuf[j*6], temp, 5);
	 sprintf (objbuf, OBJFORMAT,
		  relocatable ? RELDATA_TAG : ABSDATA_TAG, val);
	 punchrecord (outfd);
	 pc += 2;
	 j++;
	 if (j == 3)
	 {
	    printed = TRUE;
	    if (i++ == 0)
	    {
	       if (listmode && listing)
	       {
		  printline (lstfd);
	       }
	       opc = pc;
	    }
	    else
	    {
	       if (listmode && listing)
	       {
		  printdata (lstfd, opc);
	       }
	       opc = pc;
	    }
	    strcpy (opbuf, OPBLANKS);
	    j = k = 0;
	 }
      } while (term == ',');
      if (j && printed)
      {
	 printdata (lstfd, opc);
      }
      break;

   case DEF_T:
      strcpy (pcbuf, PCBLANKS);
      do {
	 bp = tokscan (bp, &tok, &toktyp, &term);
	 if (toktyp != SYM)
	 {
	    errcount++;
	    sprintf (errline[errnum++], "DEF requires symbol: %s", tok);
	 }
	 else
	 {
	    SymNode *s;

	    if (strlen (tok) > MAXSYMLEN)
	       tok[MAXSYMLEN] = '\0';

	    if ((s = symlookup (tok, FALSE)) == NULL)
	    {
	       errcount++;
	       sprintf (errline[errnum++], "DEF undefined: %s", tok);
	    }
	    else
	    {
	       s->global = TRUE;
	    }
	 }
      } while (term == ',');
      break;

   case DORG_T:
      if (!absolute && !indorg) rorgpc = pc;
      cp = exprscan (bp, &val, &term, &relocatable, 2, FALSE, 0);
      if (val < 0 || val > 65535)
      {
	 errcount++;
	 sprintf (errline[errnum++], "Invalid DORG value: %d", val);
	 val = 0;
      }
      absolute = TRUE;
      indorg = TRUE;
      pc = val;
      sprintf (pcbuf, PCFORMAT, pc);
      break;

   case DXOP_T:
      strcpy (pcbuf, PCBLANKS);
      break;

   case END_T:
      strcpy (pcbuf, PCBLANKS);
      cp = tokscan (bp, &tok, &toktyp, &term);
      if (toktyp == SYM)
      {
         SymNode *s;

	 if (strlen (tok) > MAXSYMLEN)
	 {
	    tok[MAXSYMLEN] = '\0';
	 }
	 if ((s = symlookup (tok, FALSE)) == NULL)
	 {
	    errcount++;
	    sprintf (errline[errnum++], "Undefined symbol: %s", tok);
	 }
	 else
	 {
	    sprintf (opbuf, ADDRFORMAT, s->value, s->relocatable ? '\'' : ' ');
	    sprintf (objbuf, OBJFORMAT,
		  (!absolute && s->relocatable) ? RELENTRY_TAG : ABSENTRY_TAG,
		  s->value);
	 }
      }
      return (1);
      break;

   case EQU_T:
      strcpy (pcbuf, PCBLANKS);
      if (cursym[0] == '\0')
      {
         errcount++;
	 strcpy (errline[errnum++], "EQU requires a label");
      }
      else
      {
	 SymNode *s;

	 s = symlookup (cursym, FALSE);
	 sprintf (opbuf, ADDRFORMAT, s->value & 0xFFFF,
		  s->relocatable ? '\'' : ' ');
      }
      break;

   case EVEN_T:
      opc = pc;
      if (pc & 0x0001) pc++;
      sprintf (pcbuf, PCFORMAT, pc);
      if (pc != opc)
      {
	 sprintf (objbuf, OBJFORMAT, absolute ? ABSORG_TAG : RELORG_TAG, pc);
	 punchrecord (outfd);
      }
      break;

   case IDT_T:
      strcpy (pcbuf, PCBLANKS);
      while (isspace(*bp)) bp++;
      if (*bp == '\'')
      {
	 cp = ++bp;
	 while (*bp != '\'' && *bp != '\n') bp++;
      }
      *bp = '\0';
      if (strlen(idtbuf) > IDTSIZE)
      {
	 errcount++;
	 sprintf (errline[errnum++], "Invalid IDT value: %s", cp);
	 *(cp+IDTSIZE) = '\0';
      }
      strcpy (idtbuf, cp);
      break;

   case NOP_T:
   case RT_T:
      sprintf (opbuf, OP1FORMAT, op->opmod);
      sprintf (objbuf, OBJFORMAT, ABSDATA_TAG, op->opmod);
      pc += 2;
      break;

   case LIST_T:
      strcpy (pcbuf, PCBLANKS);
      listing = TRUE;
      break;

   case OPTION_T:
      strcpy (pcbuf, PCBLANKS);
      break;

   case PAGE_T:
      strcpy (pcbuf, PCBLANKS);
      printed = TRUE;
      if (listmode && listing)
	 printline (lstfd);
      linecnt = MAXLINE;
      break;

   case REF_T:
      strcpy (pcbuf, PCBLANKS);
      break;

   case RORG_T:
      bp = exprscan (bp, &val, &term, &relocatable, 2, FALSE, 0);
      indorg = FALSE;
      absolute = FALSE;
      if (val == 0)
      {
	 pc = rorgpc;
      }
      else
      {
	 if (val < 0 || val > 65535)
	 {
	    errcount++;
	    sprintf (errline[errnum++], "Invalid RORG value: %d", val);
	    val = 0;
	 }
         pc = val;
      }
      sprintf (pcbuf, PCFORMAT, pc);
      sprintf (objbuf, OBJFORMAT, RELORG_TAG, pc);
      break;

   case TEXT_T:
      bp = tokscan (bp, &tok, &toktyp, &term);
      oval = 0;
      if (term == '-')
      {
	 oval = 1;
	 bp = tokscan (bp, &tok, &toktyp, &term);
      }
      if (toktyp == STRING)
      {
	 strcpy (opbuf, OPBLANKS);

	 bp = tok;
	 i = j = k = l = 0;
	 val = 0;
	 opc = pc;
	 while (*bp)
	 {
	    if (oval && *(bp+1) == '\0') *bp = (1 - *bp) & 0xFF;
	    byteacc = byteacc << 8 | (*bp & 0xFF);
	    sprintf (temp, CHAR1FORMAT, (*bp++ & 0xFF));
	    strncpy (&opbuf[j], temp, 2);
	    j += 2;
	    pc++;
	    bytecnt++;
	    i++;
	    if (bytecnt == 2)
	    {
	       sprintf (objbuf, OBJFORMAT, ABSDATA_TAG, byteacc);
	       punchrecord (outfd);
	       byteacc = 0;
	       bytecnt = 0;
	    }
	    if (i == 2)
	    {
	       i = 0;
	       k++;
	       if (k == 3)
	       {
		  printed = TRUE;
	          if (l++ == 0)
		  {
		     if (listmode && listing)
		     {
			printline (lstfd);
		     }
		     opc = pc;
		  }
		  else
		  {
		     if (listmode && listing)
		     {
			printdata (lstfd, opc);
		     }
		     opc = pc;
		  }
		  strcpy (opbuf, OPBLANKS);
		  j = k = 0;
	       }
	       else
	       {
		  j += 2;
	       }
	    }
	 }
	 if (k || i)
	 {
	    if (printed)
	    {
	       printdata (lstfd, opc);
	    }
	 }
      }
      else
      {
	 errcount++;
	 sprintf (errline[errnum++], "Invalid TEXT string: %s", tok);
      }
      break;

   case TITL_T:
      strcpy (pcbuf, PCBLANKS);
      while (isspace(*bp)) bp++;
      if (*bp == '\'')
      {
	 cp = ++bp;
	 while (*bp != '\'' && *bp != '\n') bp++;
      }
      *bp = '\0';
      strcpy (ttlbuf, cp);
      break;

   case UNL_T:
      strcpy (pcbuf, PCBLANKS);
      listing = FALSE;
      break;

   default: ;
   }
   return (0);
}

/***********************************************************************
* asmpass2 - Pass 2 
***********************************************************************/

int
asmpass2 (FILE *tmpfd, FILE *outfd, int listingmode, FILE *lstfd)
{
   char *bp, *cp;
   char *tok;
   int status = 0;
   int done = 0;
   int flag;
   int toktyp;
   time_t curtime;
   char term;

#ifdef DEBUG
   printf ("asmpass2: Entered\n");
#endif

   /*
   ** Get current date/time.
   */

   listmode = listingmode;
   curtime = time(NULL);
   timeblk = localtime(&curtime);
   strcpy (datebuf, ctime(&curtime));
   *strchr (datebuf, '\n') = '\0';

   /*
   ** Rewind input file.
   */

   if (fseek (tmpfd, 0, SEEK_SET) < 0)
   {
      perror ("asm990: Can't rewind temp file");
      return (-1);
   }

   /*
   ** Process the source.
   */

   memset (objrec, ' ', sizeof(objrec));

   pc = 0;
   rorgpc = 0;
   absolute = FALSE;
   indorg = FALSE;

   sprintf (objbuf, IDTFORMAT, IDT_TAG, pgmlength, idtbuf);
   punchrecord (outfd);

   byteacc = 0;
   bytecnt = 0;

   while (!done && fgets (inbuf, MAXLINE, tmpfd))
   {
#ifdef DEBUG
      printf ("in = %s", inbuf);
#endif
      linenum++;
      printed = FALSE;
      errline[0][0] = '\0';
      errnum = 0;
      objbuf[0] = '\0';
      strcpy (lstbuf, inbuf);
      bp = inbuf;

      strcpy (pcbuf, PCBLANKS);
      strcpy (opbuf, OPBLANKS);

      if (*bp != COMMENTSYM)
      {
	 OpCode *op;

	 /*
	 ** If label present, scan it off.
	 */

         if (isalpha(*bp))
	 {
	    cp = tokscan (bp, &tok, &toktyp, &term);
	    strcpy (cursym, tok);
	    if (strlen(tok) > MAXSYMLEN)
	    {
	       cursym[MAXSYMLEN] = '\0';
	       tok[MAXSYMLEN] = '\0';
	    }
	    bp = cp;
	    if (term == '\n') goto PRINTLINE;
	 }
	 else 
	 {
	    cursym[0] = '\0';
	    while (isspace (*bp)) bp++;
	 }

	 /*
	 ** Scan off opcode.
	 */

	 cp = tokscan (bp, &tok, &toktyp, &term);
	 if (!tok[0]) goto PRINTLINE;

	 /*
	 ** Process according to type.
	 */

	 if ((op = oplookup (tok)) != NULL)
	 {
	    bp = cp;

	    if (op->optype != TYPE_P)
	    {
	       if (bytecnt)
	       {
		  byteacc = (byteacc << 8) | 0;
		  sprintf (objbuf, OBJFORMAT, ABSDATA_TAG, byteacc);
		  punchrecord (outfd);
		  bytecnt = 0;
	       }
	       if (pc & 0x0001) pc++;
	    }

	    switch (op->optype)
	    {

	       case TYPE_1:
		  p21op (op, flag, bp, outfd);
	          break;

	       case TYPE_2:
		  p22op (op, flag, bp, outfd);
	          break;

	       case TYPE_3:
		  p23op (op, flag, bp, outfd);
	          break;

	       case TYPE_4:
		  p24op (op, flag, bp, outfd);
	          break;

	       case TYPE_5:
		  p25op (op, flag, bp, outfd);
	          break;

	       case TYPE_6:
		  p26op (op, flag, bp, outfd);
	          break;

	       case TYPE_7:
		  p27op (op, flag, bp, outfd);
	          break;

	       case TYPE_8:
		  p28op (op, flag, bp, outfd);
	          break;

	       case TYPE_9:
		  p29op (op, flag, bp, outfd);
	          break;

	       case TYPE_P:
	          done = p2pop (op, bp, lstfd, outfd);
	    }
	 }
	 else
	 {
	    pc++;
	    errcount++;
	    sprintf (errline[errnum++], "Invalid opcode: %s", tok);
	 }
	 
      }

      /*
      ** Write out a print line.
      */
PRINTLINE:
      if (listmode && !printed && listing)
      {
         printline (lstfd);
	 printed = FALSE;
      }

      /*
      ** Write an object buffer.
      */

      if (objbuf[0])
      {
         punchrecord (outfd);
      }

      /*
      ** Process errors
      */

      if (errnum)
      {
	 int i;

	 for (i = 0; i < errnum; i++)
	 {
	    if (listmode)
	    {
	       fprintf (lstfd, "ERROR: %s\n", errline[i]);
	       linecnt++;
	    }
	    else
	    {
	       fprintf (stderr, "asm990: %d: %s\n", linenum, errline[i]);
	    }
	    errline[i][0] = '\0';
	 }
	 status = -1;
      }

      /*
      ** Check for pass 1 error for this line
      */

      if (p1errcnt && p1error[curp1err].errorline == linenum)
      {
	 if (listmode)
	 {
	    fprintf (lstfd, "ERROR: %s\n", p1error[curp1err].errortext);
	    linecnt++;
	 }
	 else
	 {
	    fprintf (stderr, "asm990: %d: %s\n",
		     linenum, p1error[curp1err].errortext);
	 }
	 p1errcnt--;
	 curp1err++;
	 status = -1;
      }
   }

   if (!done)
   {
      errcount++;
      if (listmode)
      {
         fprintf (lstfd, "ERROR: No END record\n");
      }
      else
      {
	 fprintf (stderr, "asm990: %d: No END record\n", linenum);
      }
      status = -1;
   }
   else
   {

      /*
      ** Punch REF and DEF entries
      */

      punchsymbols (outfd);
      puncheof (outfd);

      /*
      ** Print symbol table
      */

      if (listmode)
      {
         printsymbols (lstfd);
      }
   }

   return (status);
}
